package org.simplericity.serberuhs;

import javax.security.auth.Subject;
import javax.security.auth.callback.*;
import javax.security.auth.login.AppConfigurationEntry;
import javax.security.auth.login.Configuration;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import java.util.HashMap;
import java.util.Map;
import java.io.IOException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class DefaultKerberosSubjectFactory implements KerberosSubjectFactory {

    private static Logger log = LoggerFactory.getLogger(DefaultKerberosSubjectFactory.class);
    private Subject subject;
    private KerberosSubjectConfiguration configuration;
    private static final String JAAS_ENTRYNAME = DefaultKerberosSubjectFactory.class.getName() + "_JAAS_ENTRYNAME";

    public final synchronized Subject getSubject() {
        if (configuration == null) {
            throw new NullPointerException("Configuration must be set before calling getSubject");
        }

        if (subject == null) {
            Configuration existingConfiguration = null;
            try {
                existingConfiguration = getSystemConfiguration();
            } catch (SecurityException e) {
                if (log.isDebugEnabled()) {
                    log.debug("System wide JAAS is not configured properly");
                }
            }
            try {
                Configuration.setConfiguration(createConfiguration(existingConfiguration));
                subject = createSubject();
            } finally {
                if (existingConfiguration != null) {
                    Configuration.setConfiguration(existingConfiguration);
                }
            }

        }
        return subject;
    }

    protected Configuration getSystemConfiguration() {
        return Configuration.getConfiguration();
    }

    private Subject createSubject() {
        LoginContext context;


        try {
            context = new LoginContext(JAAS_ENTRYNAME, new CallbackHandler() {
                public void handle(Callback[] callbacks) throws IOException, UnsupportedCallbackException {
                    for (Callback callback : callbacks) {
                        if(callback instanceof PasswordCallback) {
                            ((PasswordCallback)callback).setPassword(configuration.getPassword().toCharArray());
                        }
                    }
                }
            });
            context.login();

            return context.getSubject();

        } catch (LoginException e) {
            throw new RuntimeException(e);
        }
    }

    Configuration createConfiguration(final Configuration existingConfiguration) {
        Map options = new HashMap();
        options.put("useKeyTab", "true");
        options.put("keyTab", configuration.getKeytabFile().getAbsolutePath());
        options.put("storeKey", "true");
        options.put("isInitiator", "false");
        options.put("principal", configuration.getPrincipal());

        final AppConfigurationEntry entry = new AppConfigurationEntry("com.sun.security.auth.module.Krb5LoginModule", AppConfigurationEntry.LoginModuleControlFlag.REQUIRED, options);


        return new Configuration() {
            public AppConfigurationEntry[] getAppConfigurationEntry(String entryName) {
                if (JAAS_ENTRYNAME.equals(entryName)) {
                    return new AppConfigurationEntry[]{entry};
                } else {
                    Configuration parentConfiguration =
                            existingConfiguration == null ? getSystemConfiguration() : existingConfiguration;

                    return parentConfiguration.getAppConfigurationEntry(entryName);
                }
            }

            public void refresh() {

            }
        };
    }


    public final void setConfiguration(KerberosSubjectConfiguration configuration) {
        this.configuration = configuration;
    }

    public final synchronized void flushCache() {
        subject = null;
    }


}
