package org.simplericity.serberuhs;

import org.apache.commons.codec.binary.Base64;
import org.ietf.jgss.*;
import org.bouncycastle.asn1.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.security.auth.Subject;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.io.IOException;
import java.security.PrivilegedAction;
import java.util.Enumeration;
import java.nio.charset.Charset;

/**
 */
public class SpNego {
    private Logger log = LoggerFactory.getLogger(getClass());

    private SpNegoResult result;

    private String authorizedPrincipal;

    private Exception exception;

    private static final byte[] NTLMSSP_PREFIX = new byte[]{
            (byte) 'N', (byte) 'T', (byte) 'L', (byte) 'M',
            (byte) 'S', (byte) 'S', (byte) 'P', (byte) 0
    };
    private static final String KERBEROS_OID = "1.2.840.113554.1.2.2";


    public void negotiate(final Subject subject, HttpServletRequest request, HttpServletResponse response) {
        final String atz = request.getHeader("Authorization");

        if (atz == null) {
            if (log.isDebugEnabled())
                log.debug("Authorization header is missing, initiating Negotiate");
            response.setHeader("WWW-Authenticate", "Negotiate");
            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            result = SpNegoResult.MISSING_AUTHORIZATION_HEADER;
            return;
        } else if (atz.startsWith("Basic ")) {
            if (log.isDebugEnabled())
                log.debug("Client sent a Basic token instead of SPNEGO / Kerberos: " + atz);
            result = SpNegoResult.WRONG_TOKEN_BASIC;
            return;
        } else if (!isNegotiate(atz)) {
            log.debug("Client sent a WWW-Authenticate header that does not start with 'Negotiate ': " + atz);
            result = SpNegoResult.WRONG_TOKEN_NOT_NEGOTIATE;
        } else if (isNtlm(atz)) {
            log.debug("Client sent an NTLM token instead of SPNEGO / Kerberos: " + atz);
            result = SpNegoResult.WRONG_TOKEN_NTLM;
        } else {
            if (log.isDebugEnabled())
                log.debug("Authorization header is: " + atz);


            final String encodedToken = atz.substring("Negotiate ".length());

            byte[] token = Base64.decodeBase64(Charset.availableCharsets().get("utf-8").encode(encodedToken).array());

            final byte[] kerbtoken;
            try {
                kerbtoken = getKerberosToken(token);
            } catch (IOException e) {
                if (log.isDebugEnabled()) {
                    log.debug("Failed parsing token", e);
                }
                result = SpNegoResult.FAILED_PARSING_TOKEN;
                exception = e;
                return;
            }

            final String servicePrincipal = subject.getPrincipals().iterator().next().getName();

            Subject.doAs(subject, new PrivilegedAction() {
                public Object run() {

                    try {
                        final GSSManager manager = GSSManager.getInstance();



                        GSSName gssName = manager.createName(servicePrincipal, null);

                        GSSCredential serverGSSCreds = manager.createCredential(gssName,
                                GSSCredential.INDEFINITE_LIFETIME,
                                new Oid(KERBEROS_OID),
                                GSSCredential.ACCEPT_ONLY);

                        GSSContext serverGSSContext = manager.createContext(serverGSSCreds);

                        serverGSSContext.acceptSecContext(kerbtoken, 0, kerbtoken.length);

                        if (serverGSSContext.isEstablished()) {
                            log.info("Logged in " + serverGSSContext.getSrcName());
                            authorizedPrincipal = serverGSSContext.getSrcName().toString();
                        }
                        result = SpNegoResult.AUTHORIZED;
                    } catch (GSSException e) {
                        if (log.isDebugEnabled()) {
                            log.debug("SPNEGO authorization failed with a GSSException", e);
                        }
                        result = SpNegoResult.AUTHORIZATION_FAILED;
                        exception = e;
                    }
                    return null;
                }
            });
        }
    }

    private boolean isNegotiate(String atz) {
        return atz != null && atz.startsWith("Negotiate ");
    }

    byte[] getKerberosToken(byte[] token) throws IOException {
        byte[] kerbtoken = null;

            DERInputStream der = new DERInputStream(new ByteArrayInputStream(token));
            DERUnknownTag constructed = (DERUnknownTag) der.readObject();

            if (constructed.getTag() != 0x60) {
                throw new IOException("Malformed token did not start with 0x60");
            }
            InputStream tokenStream = new ByteArrayInputStream(constructed.getData());
            der = new DERInputStream(tokenStream);
            DERObjectIdentifier objectIdentifier = (DERObjectIdentifier) der.readObject();

            // This is a kerberos token
            if (KERBEROS_OID.equals(objectIdentifier.getId())) {
                kerbtoken = token;
            } else {

                ASN1TaggedObject tagged = (ASN1TaggedObject) der.readObject();

                ASN1Sequence sequence = ASN1Sequence.getInstance(tagged, true);
                Enumeration fields = sequence.getObjects();
                while (fields.hasMoreElements()) {
                    tagged = (ASN1TaggedObject) fields.nextElement();
                    if (tagged.getTagNo() == 2) {
                        ASN1OctetString mechanismToken =
                                ASN1OctetString.getInstance(tagged, true);
                        kerbtoken = mechanismToken.getOctets();
                        break;
                    }
                }
            }


        return kerbtoken;
    }

    private boolean isNtlm(String atz) {
        byte[] token = Base64.decodeBase64(Charset.availableCharsets().get("utf-8").encode(atz.substring("Negotiate ".length())).array());
        if (token.length < NTLMSSP_PREFIX.length) {
            return false;
        }
        for (int i = 0; i < NTLMSSP_PREFIX.length; i++) {
            if (token[i] != NTLMSSP_PREFIX[i]) {
                return false;
            }
        }
        return true;
    }

    public SpNegoResult getResult() {
        return result;
    }

    public Exception getException() {
        return exception;
    }

    public String getAuthorizedPrincipal() {
        return authorizedPrincipal;
    }
}